==> lhcsmapi-1.2.19.tar.gz: lhcsmapi-1.2.19/lhcsmapi/dbsignal/SignalAnalysis.py <==
import numpy as np
from scipy import signal as s
import pandas as pd
from scipy.spatial.distance import pdist, squareform
import math


class SignalAnalysis:

    @staticmethod
    def find_rows_with_exact_pattern_in_series(series, pattern):
        series_with_last_index_of_pattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        df_with_pattern = series_with_last_index_of_pattern.fillna(0)
        last_indices_of_pattern = np.where(df_with_pattern)[0]  # indices with the last found element
        all_indices_of_pattern = [range(match - len(pattern) + 1, match + 1) for match in
                               last_indices_of_pattern]  # all indices corresponding to the pattern

        data = []
        for subs in all_indices_of_pattern:
            data.append(series.iloc[subs].index)

        # flatten indices
        flattened_indices = [val for indices_of_pattern in all_indices_of_pattern for val in indices_of_pattern]

        return (pd.DataFrame(data, columns=pattern), flattened_indices)

    @staticmethod
    def resample(origin_series, ts=0, t0=None, t_end=None):
        """Function resamples series by interpolating in order to get continuous sample frequency
        """
        # takes sample time of first sample if not given
        if ts == 0:
            ts = (origin_series.index[1] - origin_series.index[0])  # Sample distance
        if t0 is not None and t_end is not None:
            index_list = SignalAnalysis._get_time_point_distribution(ts, t0, t_end)
        else:
            index_list = SignalAnalysis._get_time_point_distribution(ts, origin_series.index[0], origin_series.index[-1])

        # interpolate data in order to have the same sample time everywhere
        interpolated_series = origin_series.reindex(origin_series.index.union(index_list)).interpolate('index',
                                                                                                       limit_direction='forward')
        # delete those points which are not in in the continuous interval
        del_index = [x for x in origin_series.index if x not in index_list]
        return interpolated_series.drop(del_index, axis=0)

    @staticmethod
    def _get_time_point_distribution(ts, t0, t_end):
        t = (t_end - t0)  # Time period
        n = int(t / ts)  # Number of sample points
        return np.linspace(t0, t + t0, n)

    @staticmethod
    def apply_frequency_filter(origin_series, sample_frequency=0, cut_off_frequency=0, filter_type='low'):
        """Method applies low pass filter to series
        """
        # Set cuttoff frequency if not given
        if sample_frequency == 0:
            sample_frequency = 1 / (origin_series.index[1] - origin_series.index[0])  # Sample distance
        if cut_off_frequency == 0:
            cut_off_frequency = sample_frequency / 100  # Cut-off frequency of the filter
        normalized_frequency = cut_off_frequency / (sample_frequency / 2)  # Normalize the frequency

        b, a = s.butter(5, normalized_frequency, filter_type)
        filtered_series = pd.DataFrame(data=s.filtfilt(b, a, origin_series.values), index=origin_series.index,
                                      columns=['filtered'])
        return filtered_series['filtered']

    @staticmethod
    def normalize_df_with_global_max_and_min(df):
        df = df.dropna()
        return (df- np.min(df.values))/(np.max(df.values)-np.min(df.values))

    @staticmethod
    def normalize_d_fof_signal_group(df, signal_names):
        for signals in signal_names:
            df[signals] = SignalAnalysis.normalize_df_with_global_max_and_min(df[signals])
        return df
    
    @staticmethod
    def calculate_spectrogram_from_series(origin_series):
        """Function calculates spectrogram from DF
        """
        
        origin_series = origin_series.to_frame(name=None).set_index((origin_series.index - origin_series.index[0]))[origin_series.name]  #*10**-9
        ts = (origin_series.index[1] - origin_series.index[0]) #Sample distance
        t = (origin_series.index[-1] - origin_series.index[0]) #Timeperiod
        n = int(t/ts) # Number of samplepoints
        if len(origin_series) == n: #are the samples continuous
            new_series = origin_series
        else:
            new_series = SignalAnalysis.resample(origin_series)
        nperseg = n // 128
        try:
            f, t, sxx = s.spectrogram(new_series.values, 1/ts ,window=('hann'), nperseg=nperseg, noverlap = nperseg // 2)
            return f, t, sxx
        except Exception as e:
            print('Could not calculate spectrogram: '+ str(e)) 

    @staticmethod
    def calculate_filter_bank_from_spectrogram(spectrogram):
        """Function filterbank feature from sprectrogram
        """
        if spectrogram is not None:
            f, t, sxx = spectrogram
            filter_bank_max_values = np.zeros(len(sxx[:,1])//8)
            for i in range(len(filter_bank_max_values)):
                k = i*8           
                filter_bank_max_values[i] = np.amax(sxx[k:k+8,:])
            return np.mean(filter_bank_max_values[1:5])
        
    @staticmethod     
    def calculate_cross_correlation(ref_series, meas_series):
        """Function compares to signals and returns the correlation coefficient, the length of the signals can vary
        """
        shift = len(meas_series.values) - len(ref_series.values)
        max_correlation_timestamp = 0
        if shift > 0:
            r_val = np.zeros(abs(shift))
            i_val = np.zeros(abs(shift))

            for i in range(0 ,abs(shift)):
                i_val[i] += i
                r = np.corrcoef(meas_series.values[i:i + len(ref_series.values)], ref_series.values)
                r_val[i] = r[0][1]

            max_correlation_timestamp = meas_series.index[np.argmax(r_val)]
            max_correlation = np.max(r_val)

        elif shift < 0:
            r_val = np.zeros(abs(shift))
            i_val = np.zeros(abs(shift))

            for i in range(0 ,abs(shift)):
                i_val[i] += i
                r = np.corrcoef(ref_series.values[i:i + len(meas_series.values)], meas_series.values)
                r_val[i] = r[0][1]

            max_correlation_timestamp = ref_series.index[np.argmax(r_val)]
            max_correlation = np.max(r_val)
        else:
            r = np.corrcoef(meas_series.values, ref_series.values)
            max_correlation = r[0][1]
        return max_correlation,  max_correlation_timestamp

    @staticmethod
    def calculate_similarity_matrix(df):
        sim_m = np.full(np.math.factorial(len(df.columns) - 1), -1)
        similarity = np.nonzero(np.triu(squareform(pdist(df.T, metric='euclidean'))).flatten())[0]
        if len(similarity) > 0:
            sim_m = np.triu(squareform(pdist(df.T, metric='euclidean'))).flatten()
        return pd.DataFrame(data=sim_m[np.nonzero(sim_m)], index=np.nonzero(sim_m)[0],
                            columns=[df.columns.values[0][:-2] + '_SimilarityMatrix'])

    @staticmethod
    def compare_curve_to_curve(series_decay, series_decay_ref, envelope_amplitude,
                               envelope_characteristic_time=False, apply_filter=False, filter_cut_off_frequency=100):
        end = np.min([len(series_decay), len(series_decay_ref)])

        series_decay = series_decay.iloc[0:end]
        series_decay_ref = series_decay_ref.iloc[0:end]

        if apply_filter:
            series_decay = SignalAnalysis.apply_frequency_filter(series_decay, filter_cut_off_frequency,
                                                                 filter_type=apply_filter)
            series_decay_ref = SignalAnalysis.apply_frequency_filter(series_decay_ref, filter_cut_off_frequency,
                                                                     filter_type=apply_filter)

        index = series_decay.index
        if envelope_characteristic_time:
            envelope = np.exp(-index / envelope_characteristic_time)
            substract_array = np.abs(series_decay_ref.values - series_decay.values) / envelope
        else:
            substract_array = np.abs(series_decay_ref.values - series_decay.values)

        out_of_envelope_ratio = np.sum(substract_array > envelope_amplitude) / len(substract_array)
        return out_of_envelope_ratio

    @staticmethod
    def calculate_dtw_distance(s1, s2):
        dtw = {}

        for i in range(len(s1)):
            dtw[(i, -1)] = float('inf')
        for i in range(len(s2)):
            dtw[(-1, i)] = float('inf')
        dtw[(-1, -1)] = 0

        for i in range(len(s1)):
            for j in range(len(s2)):
                dist = (s1[i] - s2[j]) ** 2
                dtw[(i, j)] = dist + min(dtw[(i - 1, j)], dtw[(i, j - 1)], dtw[(i - 1, j - 1)])

        return math.sqrt(dtw[len(s1) - 1, len(s2) - 1])

==> lhcsmapi-1.2.111.tar.gz: lhcsmapi-1.2.111/lhcsmapi/dbsignal/SignalAnalysis.py <==
import numpy as np
from scipy import signal as s
import pandas as pd
from scipy.spatial.distance import pdist, squareform
import math


class SignalAnalysis:

    @staticmethod
    def find_rows_with_exact_pattern_in_series(series, pattern):
        series_with_last_index_of_pattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        df_with_pattern = series_with_last_index_of_pattern.fillna(0)
        last_indices_of_pattern = np.where(df_with_pattern)[0]  # indices with the last found element
        all_indices_of_pattern = [range(match - len(pattern) + 1, match + 1) for match in
                               last_indices_of_pattern]  # all indices corresponding to the pattern

        data = []
        for subs in all_indices_of_pattern:
            data.append(series.iloc[subs].index)

        # flatten indices
        flattened_indices = [val for indices_of_pattern in all_indices_of_pattern for val in indices_of_pattern]

        return (pd.DataFrame(data, columns=pattern), flattened_indices)

    @staticmethod
    def resample(origin_series, ts=0, t0=None, t_end=None):
        """Function resamples series by interpolating in order to get continuous sample frequency
        """
        # takes sample time of first sample if not given
        if ts == 0:
            ts = (origin_series.index[1] - origin_series.index[0])  # Sample distance
        if t0 is not None and t_end is not None:
            index_list = SignalAnalysis._get_time_point_distribution(ts, t0, t_end)
        else:
            index_list = SignalAnalysis._get_time_point_distribution(ts, origin_series.index[0], origin_series.index[-1])

        # interpolate data in order to have the same sample time everywhere
        interpolated_series = origin_series.reindex(origin_series.index.union(index_list)).interpolate('index',
                                                                                                       limit_direction='forward')
        # delete those points which are not in in the continuous interval
        del_index = [x for x in origin_series.index if x not in index_list]
        return interpolated_series.drop(del_index, axis=0)

    @staticmethod
    def _get_time_point_distribution(ts, t0, t_end):
        t = (t_end - t0)  # Time period
        n = int(t / ts)  # Number of sample points
        return np.linspace(t0, t + t0, n)

    @staticmethod
    def apply_frequency_filter(origin_series, sample_frequency=0, cut_off_frequency=0, filter_type='low'):
        """Method applies low pass filter to series
        """
        # Set cuttoff frequency if not given
        if sample_frequency == 0:
            sample_frequency = 1 / (origin_series.index[1] - origin_series.index[0])  # Sample distance
        if cut_off_frequency == 0:
            cut_off_frequency = sample_frequency / 100  # Cut-off frequency of the filter
        normalized_frequency = cut_off_frequency / (sample_frequency / 2)  # Normalize the frequency

        b, a = s.butter(5, normalized_frequency, filter_type)
        filtered_series = pd.DataFrame(data=s.filtfilt(b, a, origin_series.values), index=origin_series.index,
                                      columns=['filtered'])
        return filtered_series['filtered']

    @staticmethod
    def normalize_df_with_global_max_and_min(df):
        df = df.dropna()
        return (df- np.min(df.values))/(np.max(df.values)-np.min(df.values))

    @staticmethod
    def normalize_df_of_signal_group(df, signal_names):
        for signals in signal_names:
            df[signals] = SignalAnalysis.normalize_df_with_global_max_and_min(df[signals])
        return df
    
    @staticmethod
    def calculate_spectrogram_from_series(origin_series):
        """Function calculates spectrogram from DF
        """
        
        origin_series = origin_series.to_frame(name=None).set_index((origin_series.index - origin_series.index[0]))[origin_series.name]  #*10**-9
        ts = (origin_series.index[1] - origin_series.index[0]) #Sample distance
        t = (origin_series.index[-1] - origin_series.index[0]) #Timeperiod
        n = int(t/ts) # Number of samplepoints
        if len(origin_series) == n: #are the samples continuous
            new_series = origin_series
        else:
            new_series = SignalAnalysis.resample(origin_series)
        nperseg = n // 128
        try:
            f, t, sxx = s.spectrogram(new_series.values, 1/ts ,window=('hann'), nperseg=nperseg, noverlap = nperseg // 2)
            return f, t, sxx
        except Exception as e:
            print('Could not calculate spectrogram: '+ str(e)) 

    @staticmethod
    def calculate_filter_bank_from_spectrogram(spectrogram):
        """Function filterbank feature from sprectrogram
        """
        if spectrogram is not None:
            f, t, sxx = spectrogram
            filter_bank_max_values = np.zeros(len(sxx[:,1])//8)
            for i in range(len(filter_bank_max_values)):
                k = i*8           
                filter_bank_max_values[i] = np.amax(sxx[k:k+8,:])
            return np.mean(filter_bank_max_values[1:5])
        
    @staticmethod     
    def calculate_cross_correlation(ref_series, meas_series):
        """Function compares to signals and returns the correlation coefficient, the length of the signals can vary
        """
        shift = len(meas_series.values) - len(ref_series.values)
        max_correlation_timestamp = 0
        if shift > 0:
            r_val = np.zeros(abs(shift))
            i_val = np.zeros(abs(shift))

            for i in range(0 ,abs(shift)):
                i_val[i] += i
                r = np.corrcoef(meas_series.values[i:i + len(ref_series.values)], ref_series.values)
                r_val[i] = r[0][1]

            max_correlation_timestamp = meas_series.index[np.argmax(r_val)]
            max_correlation = np.max(r_val)

        elif shift < 0:
            r_val = np.zeros(abs(shift))
            i_val = np.zeros(abs(shift))

            for i in range(0 ,abs(shift)):
                i_val[i] += i
                r = np.corrcoef(ref_series.values[i:i + len(meas_series.values)], meas_series.values)
                r_val[i] = r[0][1]

            max_correlation_timestamp = ref_series.index[np.argmax(r_val)]
            max_correlation = np.max(r_val)
        else:
            r = np.corrcoef(meas_series.values, ref_series.values)
            max_correlation = r[0][1]
        return max_correlation,  max_correlation_timestamp

    @staticmethod
    def calculate_similarity_matrix(df):
        sim_m = np.full(np.math.factorial(len(df.columns) - 1), -1)
        similarity = np.nonzero(np.triu(squareform(pdist(df.T, metric='euclidean'))).flatten())[0]
        if len(similarity) > 0:
            sim_m = np.triu(squareform(pdist(df.T, metric='euclidean'))).flatten()
        return pd.DataFrame(data=sim_m[np.nonzero(sim_m)], index=np.nonzero(sim_m)[0],
                            columns=[df.columns.values[0][:-2] + '_SimilarityMatrix'])

    @staticmethod
    def compare_curve_to_curve(series_decay, series_decay_ref, envelope_amplitude,
                               envelope_characteristic_time=False, apply_filter=False, filter_cut_off_frequency=100):
        end = np.min([len(series_decay), len(series_decay_ref)])

        series_decay = series_decay.iloc[0:end]
        series_decay_ref = series_decay_ref.iloc[0:end]

        if apply_filter:
            series_decay = SignalAnalysis.apply_frequency_filter(series_decay, filter_cut_off_frequency,
                                                                 filter_type=apply_filter)
            series_decay_ref = SignalAnalysis.apply_frequency_filter(series_decay_ref, filter_cut_off_frequency,
                                                                     filter_type=apply_filter)

        index = series_decay.index
        if envelope_characteristic_time:
            envelope = np.exp(-index / envelope_characteristic_time)
            substract_array = np.abs(series_decay_ref.values - series_decay.values) / envelope
        else:
            substract_array = np.abs(series_decay_ref.values - series_decay.values)

        out_of_envelope_ratio = np.sum(substract_array > envelope_amplitude) / len(substract_array)
        return out_of_envelope_ratio

    @staticmethod
    def calculate_dtw_distance(s1, s2):
        dtw = {}

        for i in range(len(s1)):
            dtw[(i, -1)] = float('inf')
        for i in range(len(s2)):
            dtw[(-1, i)] = float('inf')
        dtw[(-1, -1)] = 0

        for i in range(len(s1)):
            for j in range(len(s2)):
                dist = (s1[i] - s2[j]) ** 2
                dtw[(i, j)] = dist + min(dtw[(i - 1, j)], dtw[(i, j - 1)], dtw[(i - 1, j - 1)])

        return math.sqrt(dtw[len(s1) - 1, len(s2) - 1])
